import time
import json
import logging
import io
import os
import yaml
from flask import Flask, request
from .bridge import bridge_adapter_factory
from .proxy.mqtt import MqttProxy
from .device import DeviceManager

app = Flask(__name__)
bridge = None
proxy = None
device_manager = None
configuration = None

def init(config_path=None, default_config=None):
    global bridge, proxy, device_manager, configuration
    configuration = _load_config(config_path,default_config)
    device_manager = DeviceManager(configuration['storage'])
    bridge = bridge_adapter_factory.create(device_manager,configuration['bridge'],
        on_config_handler=_on_config,
        on_commands_handler=_on_commands)
    if 'mqtt' in configuration:
        proxy=MqttProxy(configuration['mqtt'],bridge)
        proxy.start()
        logging.debug("MQTT proxy is enable: {}".format(proxy.is_ready()))
    else:
        logging.debug('MQTT proxy is disabled')
    bridge.connect()
    return app

@app.route('/',methods = ['GET'])
def index():
    return 'OK'

@app.route('/device/<device_id>',methods = ['GET'])
def get_device(device_id):
    device = device_manager.get_device(device_id)
    return json.dumps(device.toJson())

def _load_config(config_path='/etc/iot-gw/configuration.yml',default_config=None):
    if config_path is None or not os.path.isfile(config_path):
        result = default_config
    else:
        with io.open(config_path,'r') as stream:
            result = yaml.safe_load(stream)
    return result

def _on_config(device_id,configuration):
    global proxy
    proxy.config(device_id,configuration)

def _on_commands(device_id,commands):
    global proxy
    proxy.commands(device_id,commands)





import time
import logging
from iot_gw.proxy.mqtt import MqttProxy


logging.basicConfig(level=logging.DEBUG)
proxy=MqttProxy('test','gateway','P@ssw0rd','./tests/data/ca.crt',
    on_attach=lambda d:print("on_attach {}".format(d)),
    on_unattach=lambda d:print("on_unattach {}".format(d)),
    on_event=lambda d,p:print("on_event {} / {}".format(d,p)),
    on_state=lambda d,p:print("on_sate {} / {}".format(d,p))
)
is_connected=proxy.connect(hostname='192.168.99.11',port=8883,async_connect=True)
print(is_connected)

while proxy.is_connected():
    time.sleep(1)


import unittest
import os
import iot_gw.gateway as gw

dir_path = os.path.dirname(os.path.realpath(__file__))

class GatewayTest(unittest.TestCase):

    def test_load_default_config(self):
        default_config = {
            'storage':{
                'key_pair_path' : './tests/data'
            },
            'bridge':{
                'adapter': 'gcp',
                'project_id': 'iot-dev-260617',
                'region': 'europe-west1',
                'registry_id': 'mylab',
                'device_id': 'gw-dev',
                'private_key_file': './tests/data/gw_private.pem',
                'ca_certs_file': './tests/data/mqtt.googleapis.com.pem',
                'bridge_hostname': 'mqtt.googleapis.com',
                'bridge_port': 443
            },
            'server':{
                'http':{
                    'host':'0.0.0.0',
                    'port':'8080'
                }
            }
        }
        # when config_path is undefined, use default config
        config = gw._load_config(config_path = None, default_config=default_config)
        self.assertEquals(default_config,config)
        # when file is not found, use default config
        config = gw._load_config(config_path="/wrong/path.yml",default_config=default_config)
        self.assertEquals(default_config,config)

    def test_load_config(self):
        config = gw._load_config(config_path=os.path.join(dir_path,"data/configuration.yml"),default_config=None)
        self.assertIsNotNone(config)
        self.assertTrue('storage' in config)

        




import time
import ssl
import logging
from enum import Enum
import paho
from .adapter import ProxyAdapter

class MqttProxy(ProxyAdapter):
    def __init__(self,config,adapter):
        super().__init__('mqtt',adapter)
        self.__client_id='mqtt_proxy'
        self.__username=config['login']
        self.__password=config['password']
        self.__hostname=config['hostname']
        self.__port=config['port']
        self.__ca_certs_file = config['ca_certs_file'] if 'ca_certs_file' in config else None
        self.__client=paho.mqtt.client.Client(client_id=self.__client_id)
        self.__client.username_pw_set(username=self.__username,password=self.__password)
        if not self.__ca_certs_file is None:
            logging.debug("Enable TLS")
            self.__client.tls_set(self.__ca_certs_file,tls_version= ssl.PROTOCOL_TLSv1_2)
        self.__client.enable_logger(logging)
        self.__client.on_connect=self.__on_connect
        self.__client.on_disconnect=self.__on_disconnect
        self.__client.on_message=self.on_message
        self.__is_connected=False
        self.__topicHandlers={
            'attach' : self.__on_attach_message,
            'unattach' : self.__on_unattach_message,
            'state' : self.__on_state_message,
            'event' : self.__on_event_message
        }
        
    def start(self):
        return self.connect()

    def connect(self,timeout=5,async_connect=True):
        if async_connect:
            self.__client.connect_async(self.__hostname,int(self.__port))
            self.__client.loop_start()
            self.__wait_for_connection(timeout)
        else:
            self.__client.connect(self.__hostname,self.__port)
        return self.is_connected()

    def is_ready(self):
        return self.is_connected()

    def is_connected(self):
        return self.__is_connected

    def config(self,device_id,configuration):
        logging.debug("Config device {}: {}".format(device_id,configuration))
        self.__client.publish(
            topic="/config/{}".format(device_id),
            payload=configuration,
            qos=0)

    def commands(self,device_id,command):
        logging.debug("Commands device {}: {}".format(device_id,command))
        self.__client.publish(
            topic="/commands/{}".format(device_id),
            payload=command,
            qos=0)

    def on_message(self,client,userdata,message):
        payload = str(message.payload.decode('utf-8'))
        logging.debug(
            'Received message \'{}\' on topic \'{}\' with Qos {}'
            .format(payload, message.topic, str(message.qos))
        )
        topics= list(filter(lambda t : len(t) > 0,message.topic.split('/')))
        if topics[0] in self.__topicHandlers:
            self.__topicHandlers[topics[0]](message.payload,topics[1:])
        else:
            logging.warn("MQTT proxy have receiced message on an unknown topic: %s",topics[0])

    def __on_attach_message(self,payload,subtopics):
        device_id=payload.decode('utf-8')
        self.__client.subscribe('/event/{}'.format(device_id))
        self.__client.subscribe('/state/{}'.format(device_id))
        self.attach_handler(device_id)

    def __on_unattach_message(self,payload,subtopics):
        device_id=payload.decode('utf-8')
        self.__client.unsubscribe('/event/{}'.format(device_id))
        self.__client.unsubscribe('/state/{}'.format(device_id))
        self.unattach_handler(device_id)

    def __on_event_message(self,payload,subtopics):
        self.event_handler(subtopics[0],payload)

    def __on_state_message(self,payload,subtopics):
        self.state_handler(subtopics[0],payload)

    def __on_connect(self,client,userdata,flags,rc):
        logging.debug("MQTT client %s connection is up" % self.__client_id)
        self.__client.subscribe('/attach',qos=1)
        self.__client.subscribe('/unattach',qos=1)
        self.__is_connected=True

    def __on_disconnect(self,client,userdate,rc):
        logging.debug("MQTT client %s connection is down" % self.__client_id)
        self.__is_connected=False


    def __wait_for_connection(self, timeout=5):
        total_time = 0
        while not self.__is_connected and total_time < timeout:
            logging.debug("wait_for_connection %d" % timeout)
            if timeout > 0:
                total_time +=1
            time.sleep(1)
        if not self.__is_connected:
            raise RuntimeError('Could not connect to MQTT server.')
        logging.debug("wait_for_connection terminated %s" % self.is_connected())




import logging
from .gcp import MqttBridge as GCPMqttBridge

class BridgeAdapterFactory:
    """
    BridgeAdapter factory

    The factory instanciate a bridge adapter based on adapter property

    Supported adapter type:
      - gcp: Google Cloud Platform mqtt bridge
      - gcp: Google Cloud Platform mqtt bridge
    """
    
    def __init__(self):
        self.__adapters={
            'gcp': GCPMqttBridge,
            'gcp_mqtt': GCPMqttBridge
        }

    def create(self,device_manager,config,on_config_handler=None,on_commands_handler=None):
        """
        Create a BridgeAdapter instance

        Parameters
        ----------
        device_manager : DeviceManager
            Device manager component

        config : dict
            dictionnary with properties which are used to establish the connection.
        
        on_config_handler : method, optional
            Method which treats configuration message received from bridge

        on_commands_handler : method, optional
            Method which treats commands message received from bridge

        Returns
        -------
            Bridge adapter instance

        Raises
        ------
            RuntimeError if adapter property is not defined in config 
            RuntimeError if adapter property value is not supported
        """
        try:
            adapter_id=config['adapter']
        except KeyError:
            raise RuntimeError("Undefined bridge adapter in configuration")
        adapter = None
        try:
            adapter = self.__adapters[adapter_id](device_manager,config,
                on_config=on_config_handler,
                on_commands=on_commands_handler)
        except KeyError:
            raise RuntimeError("Unknown bridge adapter: %s" % adapter_id)
        return adapter




class BridgeAdapter:
    """
        Bridge Adapter interface
    """

    def __init__(self,adapter,device_manager):
        self.__adapter=adapter
        self.__device_manager=device_manager

    def get_adapter(self):
        return self.__adapter

    def get_device_manager(self):
        return self.__device_manager

    def connect(self):
        """
        Establish connection to the bridge

        Returns
        -------
            True or False
        """
        raise NotImplementedError("Not implemented method connect for bridge adapter: %s" % self.get_adapter())

    def attach(self,device_id):
        """
        Attach a device to the gateway

        Parameters
        ----------
        device_id : str 
            device identifiant

        Returns
        -------
            True or False
        """
        raise NotImplementedError("Not implemented method attach for bridge adapter: %s" % self.get_adapter())

    def unattach(self,device_id):
        """
        Unattach a device from the gateway

        Parameters
        ----------
        device_id : str
            device identifiant

        Returns
        -------
            True or False
        """
        raise NotImplementedError("Not implemented method unattach for bridge adapter: %s" % self._adapter)

    def publish_event(self,payload,device_id,qos):
        """
        Publish an event message to bridge from a device

        Parameters
        ----------
        payload : str
            message payload
        device_id: str
            device identifiant
        qos: int, optional
            quality of service. 

        Returns
        -------
            True if message is published
            False if message is not published
        """
        raise NotImplementedError("Not implemented method publish_event for bridge adapter: %s" % self._adapter)

    def publish_state(self,payload,device_id,qos):
        """
        Publish a state message to bridge from a device

        Parameters
        ----------
        payload : str
            message payload
        device_id: str
            device identifiant
        qos: int, optional
            quality of service. 

        Returns
        -------
            True if message is published
            False if message is not published
        """
        raise NotImplementedError("Not implemented method publish_state for bridge adapter: %s" % self._adapter)

    def is_connected(self):
        """
        Get connection status

        Returns
        -------
            True if connection is established
            False if connection is not established
        """
        raise NotImplementedError("Not implemented method is_connected for bridge adapter: %s" % self._adapter)



import time
import os
import unittest
import shutil
import jwt
from iot_gw.device import Device
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.backends import default_backend as crypto_default_backend


temp_dir="./.workdir"

class TestDevice(unittest.TestCase):

    def setUp(self):
        #create temporary folder
        shutil.rmtree(temp_dir,ignore_errors=True)
        os.mkdir(temp_dir)

    def tearDown(self):
        shutil.rmtree(temp_dir,ignore_errors=True)
    

    def test_dump(self):
        device = Device('test_dump')
        device.dump(temp_dir)
        for file in ['test_dump_private.pem','test_dump_public.pem']:
            self.assertTrue(os.path.isfile(os.path.join(temp_dir,file)))

    def test_fetch(self):
        key = rsa.generate_private_key(
            backend=crypto_default_backend(),
            public_exponent=65537,
            key_size=2048
        )
        public_key = key.public_key().public_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PublicFormat.PKCS1
        )
        private_key = key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.PKCS8,
            crypto_serialization.NoEncryption()
        )
        with open(os.path.join(temp_dir,'test_fetch_private.pem'),'wb') as private_key_file:
            private_key_file.write(private_key)
        with open(os.path.join(temp_dir,'test_fetch_public.pem'),'wb') as public_key_file:
            public_key_file.write(public_key)
        device = Device('test_fetch')
        device.fetch(temp_dir)
        self.assertEqual(public_key,device.get_public_key())
        self.assertEqual(private_key,device.get_private_key())





        

if __name__ == '__main__':
    unittest.main()



import unittest
import jwt
import logging
from iot_gw.bridge.gcp import create_jwt_token, create_mqtt_client, MqttBridge
import os 
dir_path = os.path.dirname(os.path.realpath(__file__))
gw_private_key = os.path.join(dir_path,'data/gw_private.pem')
gw_public_key = os.path.join(dir_path,'data/gw_public.pem')
ca_certs = os.path.join(dir_path, 'data/mqtt.googleapis.com.pem')

mqtt_bridge_config={
    'project_id': 'iot-dev-260617',
    'region': 'europe-west1',
    'registry_id': 'mylab',
    'device_id': 'gw-dev',
    'private_key_file': gw_private_key,
    'ca_certs_file': ca_certs,
    'bridge_hostname': 'mqtt.googleapis.com',
    'bridge_port': 443
}

class TestGcpBridge(unittest.TestCase):
    
    # def setUp(self):
    #     logging.basicConfig(level=logging.DEBUG)
        
    def test_create_jwt_token(self):
        token=create_jwt_token('project_id',gw_private_key)
        self.assertIsNotNone(token)
        with open(gw_public_key, 'r') as f:
            public_key = f.read()  
        decodedToken=jwt.decode(token,public_key,algorithms='RS256',options={'verify_aud': False})
        self.assertEqual(decodedToken['aud'],'project_id')
        self.assertEqual(decodedToken['exp'] - decodedToken['iat'], 3600)

    def test_create_mqtt_client(self):
        client=create_mqtt_client(
            'project_id',
            'europe-west-1',
            'registry_id',
            'device_id',
            gw_private_key,
            ca_certs)
        self.assertIsNotNone(client)

    def test_connect_mqtt_bridge(self):
        bridge = MqttBridge(None,mqtt_bridge_config)
        try:
            bridge.connect()
        except RuntimeError:
            self.fail()
        self.assertTrue(bridge.is_connected())



if __name__ == '__main__':
    unittest.main()


import datetime
import os
import enum
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.backends import default_backend as crypto_default_backend
import jwt

class DeviceStatus(enum.Enum):
    UNKNOWN = 0
    REGISTERED = 1
    ATTACHED = 2


class Device:
    def __init__(self,device_id,key_folder=None):
        self.device_id = device_id
        self.status = DeviceStatus.UNKNOWN
        if not self.__key_pair_is_available(key_folder):
            self.__generate_key_pair()
        else:
            self.fetch(key_folder)

    def toJson(self):
        return {
            "device_id" : self.device_id,
            "public_key" : self.__public_key.decode('utf-8')
        }

    def get_public_key(self):
        return self.__public_key

    def get_private_key(self):
        return self.__private_key

    def dump(self, path):
        self.__dump(
            self.get_private_key(),
            self.__build_private_key_path(path)
        )
        self.__dump(
            self.get_public_key(),
            self.__build_public_key_path(path)
        )

    def fetch(self, path):
        with open(self.__build_private_key_path(path),'rb') as private_key_file:
            self.__private_key = crypto_serialization.load_pem_private_key(
                private_key_file.read(),
                password=None,
                backend=crypto_default_backend()
            ).private_bytes(
                crypto_serialization.Encoding.PEM,
                crypto_serialization.PrivateFormat.PKCS8,
                crypto_serialization.NoEncryption()
            )
        with open(self.__build_public_key_path(path),'rb') as public_key_file:
            self.__public_key = crypto_serialization.load_pem_public_key(
                public_key_file.read(),
                backend=crypto_default_backend()
            ).public_bytes(
                crypto_serialization.Encoding.PEM,
                crypto_serialization.PublicFormat.PKCS1
            ) 
        
    def __generate_key_pair(self):
        key = rsa.generate_private_key(
            backend=crypto_default_backend(),
            public_exponent=65537,
            key_size=2048
        )
        self.__public_key = key.public_key().public_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PublicFormat.PKCS1
        )
        self.__private_key = key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.PKCS8,
            crypto_serialization.NoEncryption()
        )

    def __build_private_key_path(self,path):
        return os.path.join(path,'{}_private.pem'.format(self.device_id))

    def __build_public_key_path(self,path):
        return os.path.join(path,'{}_public.pem'.format(self.device_id))
        
    def __dump(self,data,file):
        with open(file,'wb') as output:
            output.write(data)
    
    def __key_pair_is_available(self,path):
        if path is None:
            return False
        else:
            public_key_path = self.__build_public_key_path(path)
            private_key_path = self.__build_private_key_path(path)
            return os.path.isfile(self.__build_public_key_path(path)) and os.path.isfile(private_key_path)


class DeviceManager:

    def __init__(self,config):
        self.__key_pair_path = config['key_pair_path']

    def get_device(self,device_id):
        return Device(device_id,self.__key_pair_path)


    


import unittest
import paho.mqtt.client
from iot_gw.proxy.mqtt import MqttProxy
from mockito import mock, verify, when


mqtt_config={
    'login':'username',
    'password':'password',
    'hostname':'localhost',
    'port': 1886
}

class Message:
    def __init__(self,topic,payload,qos=0):
        self.topic=topic
        self.payload=payload.encode('utf-8')
        self.qos=qos

class MqttProxyTest(unittest.TestCase):

    def test_attach_topic(self):
        mqttMock = mock()
        adapterMock = mock()
        when(paho.mqtt.client).Client(client_id='mqtt_proxy').thenReturn(mqttMock)
        proxy = MqttProxy(mqtt_config,adapterMock)
        message = Message('/attach','device_id')
        proxy.on_message(client='client_id',userdata=None,message=message)
        verify(adapterMock,times=1).attach('device_id')
        verify(mqttMock,times=1).subscribe('/event/device_id')
        verify(mqttMock,times=1).subscribe('/state/device_id')

    def test_unattach_topic(self):
        adapterMock = mock()
        mqttMock = mock()
        when(paho.mqtt.client).Client(client_id='mqtt_proxy').thenReturn(mqttMock)
        proxy = MqttProxy(mqtt_config,adapterMock)
        message = Message('/unattach','device_id')
        proxy.on_message(client='client_id',userdata=None,message=message)
        verify(adapterMock,times=1).unattach('device_id')
        verify(mqttMock,times=1).unsubscribe('/event/device_id')
        verify(mqttMock,times=1).unsubscribe('/state/device_id')

    def test_config(self):
        mqttMock = mock()
        when(paho.mqtt.client).Client(client_id='mqtt_proxy').thenReturn(mqttMock)
        proxy = MqttProxy(mqtt_config,mock())
        proxy.config('device_id','configuration')
        verify(mqttMock,times=1).publish(
            topic='/config/device_id',
            payload='configuration',
            qos=0
        )
    
    def test_commands(self):
        mqttMock = mock()
        when(paho.mqtt.client).Client(client_id='mqtt_proxy').thenReturn(mqttMock)
        proxy = MqttProxy(mqtt_config,mock())
        proxy.commands('device_id','command')
        verify(mqttMock,times=1).publish(
            topic='/commands/device_id',
            payload='command',
            qos=0
        )

    def test_state_topic(self):
        listenerMock = mock()
        proxy = MqttProxy(mqtt_config,listenerMock)
        message = Message('/state/device_id','state')
        proxy.on_message(client='client_id',userdata=None,message=message)
        verify(listenerMock,times=1).publish_state(b'state','device_id')

    def test_event_topic(self):
        listenerMock = mock()
        proxy = MqttProxy(mqtt_config,listenerMock)
        message = Message('/event/device_id','event')
        proxy.on_message(client='client_id',userdata=None,message=message)
        verify(listenerMock,times=1).publish_event(b'event','device_id')



    



import unittest
from iot_gw.bridge import bridge_adapter_factory;


class BridgeFactoryTest(unittest.TestCase):

    def test_gcp_mqtt_bridge(self):
        config = {
                'adapter': 'gcp',
                'project_id': 'iot-dev-260617',
                'region': 'europe-west1',
                'registry_id': 'mylab',
                'device_id': 'gw-dev',
                'private_key_file': './tests/data/gw_private.pem',
                'ca_certs_file': './tests/data/mqtt.googleapis.com.pem',
                'bridge_hostname': 'mqtt.googleapis.com',
                'bridge_port': 443
        }
        adapter = bridge_adapter_factory.create(None,config)
        self.assertIsNotNone(adapter)

    def test_missing_adapter_property(self):
        self.assertRaises(RuntimeError,bridge_adapter_factory.create,None,{})

    def test_unknown_adapter(self):
        self.assertRaises(RuntimeError,bridge_adapter_factory.create,None,{'adapter':'unknown'})    



import setuptools

with open('readme.md','r') as fh:
    readme = fh.read()

def local_scheme(version):
    return version.format_choice("","")


setuptools.setup(
    name="iot_gw",
    use_scm_version={
        "root": ".", 
        "relative_to": __file__,
        "local_scheme": local_scheme
    },
    author="A. LE CANN",
    author_email="arnaud@lecann.com",
    description="IoT gateway",
    long_description = readme,
    long_description_content_type="text/markdown",
    url="https://github.com/arnlec/iot-gw",
    packages=setuptools.find_packages(),
    install_requires=[
        'pyjwt==1.7.1',
        'cryptography==2.7',
        'flask==1.1.1',
        'paho-mqtt==1.4.0',
        'PyYAML==5.2'
    ],
    setup_requires=['setuptools_scm'],
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ],
    python_requires='>=3'


)


class ProxyAdapter:

    """
    Proxy Adapter interface
    """

    def __init__(self,id,adapter):
        self.__adapter_id=id
        self.__adapter=adapter

    def start(self):
        """
        Start proxy

        Returns
        -------
            True or False
        """
        raise NotImplementedError("Not implemented method connect for proxy adapter: %s" % self.__adapter_id)

    def is_ready(self):
        """
        Get proxy status

        Returns
        -------
            True or False
        """
        raise NotImplementedError("Not implemented method is_ready for proxy adapter: %s" % self.__adapter_id)

    def commands(self,device_id,commands):
        """
        Send commands message to a device

        Parameters
        ----------
        device_id: str
            Device identifiant

        commands: str
            commands message

        Returns
        -------
            True or False
        """
        raise NotImplementedError("Not implemented method commands for proxy adapter: %s" % self.__adapter_id)

    def config(self,device_id,configuration):
        """
        Send configuration message

        Parameters
        ----------
        device_id: str
            Device identifiant

        configuration: str
            configuration message

        Returns
        -------
            True or False
        """
        raise NotImplementedError("Not implemented method config for proxy adapter: %s" % self.__adapter_id)


    def attach_handler(self,device_id):
        self.__adapter.attach(device_id)

    def unattach_handler(self,device_id):
        self.__adapter.unattach(device_id)
        
    def event_handler(self,device_id,event):
        self.__adapter.publish_event(event,device_id)
        
    def state_handler(self,device_id,event):
        self.__adapter.publish_state(event,device_id)




import time
import datetime
import json
import jwt
import ssl
import paho.mqtt.client as mqtt
import logging
from .adapter import BridgeAdapter

"""
Bridge client for Google Cloud Platform
"""

def create_jwt_token(project_id,private_key_file,algorithm='RS256',minutes=60):
    """
    Create a JWT Token

    JWT token are use by MQTT bridge for device authentication
    """
    token = {
            'iat': datetime.datetime.utcnow(),
            'exp': datetime.datetime.utcnow() + datetime.timedelta(minutes=minutes),
            'aud' : project_id
        }
    with open(private_key_file, 'r') as f:
        private_key = f.read()     
    return jwt.encode(token,private_key,algorithm=algorithm)

def create_mqtt_client(project_id,region,registry_id, device_id, private_key_file, ca_certs_file):
    client_id='projects/{}/locations/{}/registries/{}/devices/{}'.format(
            project_id,
            region,
            registry_id,
            device_id
        )
    client = mqtt.Client(client_id=client_id)
    client.username_pw_set(
        username = 'unused',
        password = create_jwt_token(project_id,private_key_file)
    )
    client.enable_logger(logging)
    client.tls_set(ca_certs =ca_certs_file, tls_version= ssl.PROTOCOL_TLSv1_2)
    return client

class MqttBridge(BridgeAdapter):
    """
    MQTT bridge client for Google Cloud Platform 

    Configuration properties:
      - project_id
      - region
      - registry_id
      - device_id
      - private_key_file
      - ca_certs_file
      - bridge_hostname
      - bridge_port
    """
    def __init__(self,device_manager,config,on_config=None,on_commands=None):
        super().__init__('gcp_mqtt',device_manager)
        self.__is_connected=False
        self.__config=config
        self.__client=create_mqtt_client(
            config['project_id'],
            config['region'],
            config['registry_id'],
            config['device_id'],
            config['private_key_file'],
            config['ca_certs_file']
        )
        self.__client.on_connect = self.__on_connect
        self.__client.on_disconnect = self.__on_disconnect
        self.__client.on_message = self.__on_message
        self.__topic_handlers = {
            'config' : on_config,
            'commands' : on_commands,
            'errors' : self.__on_errors_handler
        }
    
    def connect(self):
        self.__client.connect_async(
            self.__config['bridge_hostname'],
            self.__config['bridge_port']
        )
        self.__client.loop_start()
        self.__wait_for_connection(timeout=5)

    def reconnect(self):
        logging.debug("Trying reconnect MQTT bridge")
        self.__client.username_pw_set(
            username = 'unused',
            password = create_jwt_token(self.__config['project_id'],self.__config['private_key_file'])
        )
        self.connect()

    def attach(self,device_id):
        jwt_token = self.__get_jwt_token(device_id)
        payload = json.dumps({"authorization" : jwt_token.decode('utf-8')}) if jwt_token is not None else None
        attached = self.__publish(payload,device_id,'attach',1)
        self.__client.subscribe('/devices/{}/config'.format(device_id), qos=1)
        self.__client.subscribe('/devices/{}/commands/#'.format(device_id), qos=1)
        self.__client.subscribe('/devices/{}/errors'.format(device_id), qos=0)
        return attached

    def unattach(self,device_id):
        jwt_token = self.__get_jwt_token(device_id)
        payload = json.dumps({"authorization" : jwt_token.decode('utf-8')}) if jwt_token is not None else None
        unattached = self.__publish(payload,device_id,'unattach',1)  
        self.__client.unsubscribe('/devices/{}/config'.format(device_id))
        self.__client.unsubscribe('/devices/{}/commands/#'.format(device_id))
        self.__client.unsubscribe('/devices/{}/errors'.format(device_id)) 
        return unattached 
            
    def publish_event(self,payload,device_id=None,qos=0):
        self.__publish(payload,device_id,'events',qos)

    def publish_state(self,payload,device_id=None,qos=0):
        self.__publish(payload,device_id,'state',qos) 

    def __publish(self,payload,device_id=None,type='events',qos=0):
        if not self.__is_connected:
            self.connect()
        result = self.__client.publish(
            topic="/devices/{}/{}".format(device_id if device_id != None else self.config['device_id'],type),
            payload=payload,
            qos=qos) 
        result.wait_for_publish()
        return result.is_published()

    def is_connected(self):
        return self.__is_connected

    def __wait_for_connection(self, timeout=5):
        total_time = 0
        while not self.__is_connected and total_time < timeout:
            logging.debug("wait_for_connection %d" % timeout)
            if timeout > 0:
                total_time +=1
            time.sleep(1)
        if not self.__is_connected:
            raise RuntimeError('Could not connect to MQTT bridge.')
        logging.debug("wait_for_connection terminated %s" % self.is_connected())
        
    def __on_connect(self,client,userdata,flags,rc):
        logging.debug("MQTT bridge connection is up")
        self.__is_connected=True
        self.__client.subscribe('/devices/{}/config'.format(self.__config['device_id']), qos=1)
        self.__client.subscribe('/devices/{}/errors'.format(self.__config['device_id']), qos=0)

    def __on_disconnect(self,client,userdate,rc):
        logging.debug("MQTT bridge connection is down")
        self.__is_connected=False
        self.reconnect()

    def __on_message(self, client, userdata, message):
        payload = str(message.payload.decode('utf-8'))
        logging.debug(
            'Received message \'{}\' on topic \'{}\' with Qos {}'
            .format(payload, message.topic, str(message.qos))
        )
        topics= list(filter(lambda t : len(t) > 0,message.topic.split('/')))
        device_id = topics[1]
        topic = topics[2]
        if topic in self.__topic_handlers:
            handler =self.__topic_handlers[topic]
            if not handler is None:
                handler(device_id,message.payload)

    def __on_errors_handler(self,device_id,payload):
        pass

    def __get_jwt_token(self,device_id):
        device = self.get_device_manager().get_device(device_id)
        token = {
            'iat': datetime.datetime.utcnow(),
            'exp': datetime.datetime.utcnow() + datetime.timedelta(minutes=60,seconds=0),
            'aud' : self.__config['project_id']
        }
        return jwt.encode(
            token,
            device.get_private_key(),
            algorithm='RS256')
        

    
    
